local Castbars = LibStub("AceAddon-3.0"):NewAddon("Castbars");

Castbars.SharedMedia = LibStub("LibSharedMedia-3.0");
Castbars.DoNothing = function() end;

function Castbars:GcdSpellNameFind()
    local spells = {1126, 1494, 168, 465, 1243, 1752, 8017, 687, 49895, 6673};
    for i, spellId in ipairs(spells) do
        local spellName = GetSpellInfo(spellId);
        if (GetSpellCooldown(spellName)) then
            self.GcdSpellNameFind = nil;
            return spellName;
        end
    end
end

function Castbars:FrameMediaRestore(frame)
    local barTexture = self.SharedMedia:Fetch("statusbar", self.db.profile[frame.configName]["Texture"]);
    local borderTexture = self.SharedMedia:Fetch("border", self.db.profile[frame.configName]["Border"]);
    local font = self.SharedMedia:Fetch("font", self.db.profile[frame.configName]["Font"]);
    if (barTexture) then
        frame.statusBar:SetStatusBarTexture(barTexture);
    end
    if (borderTexture) then
        frame.backdrop:SetBackdrop({edgeFile = borderTexture, edgeSize = 16, tileSize = 16, insets = {bottom = 5, top = 5, right = 5, left = 5}});
        local brightness = self.db.profile[frame.configName]["BorderBrightness"];
        frame.backdrop:SetBackdropBorderColor(brightness, brightness, brightness, 1.0);
    end
    if (font) then
        local textSize = self.db.profile[frame.configName]["FontSize"];
        local outline = self.db.profile[frame.configName]["FontOutline"] and "OUTLINE" or nil;
        frame.text:SetFont(font, textSize, outline);
        if (frame.timer) then
            frame.timer:SetFont(font, textSize, outline);
        end
    end
end

function Castbars:FrameMediaRestoreAll()
    for i, frame in pairs(self.frames) do
        self:FrameMediaRestore(frame)
    end
end

function Castbars:FrameRestoreColor(frame)
    if (frame.shield and frame.shield:IsShown() or frame.outOfRange) then
        frame.statusBar:SetStatusBarColor(0.6, 0.6, 0.6);
    elseif (self.db.profile[frame.configName]["UseBarColor"]) then
        frame.statusBar:SetStatusBarColor(unpack(self.db.profile[frame.configName]["BarColor"]));
    end
end

function Castbars:FrameRestoreOnCast(frame, unit)
    self:FrameRestoreColor(frame);
    if (frame.icon) then
        if (frame.shield and frame.shield:IsShown() and (self.db.profile[frame.configName]["ShowShield"])) then
            local hb, wb = frame.shield:GetHeight(), frame.shield:GetWidth();
            frame.icon:ClearAllPoints();
            frame.icon:SetPoint("TOPRIGHT", frame, "TOPLEFT", -0.03515625 * wb, 0.09375 * hb);
            frame.icon:SetHeight(0.328125 * hb);
            frame.icon:SetWidth(0.08203125 * wb);
            frame.icon:Show();
        elseif (self.db.profile[frame.configName]["ShowIcon"]) then
            frame.icon:ClearAllPoints();
            if (self.db.profile[frame.configName]["Border"] == "None") then
                frame.icon:SetPoint("RIGHT", frame, "LEFT", 0, 0);
                frame.icon:SetHeight(self.db.profile[frame.configName]["Height"]);
                frame.icon:SetWidth(self.db.profile[frame.configName]["Height"]);
            else
                frame.icon:SetPoint("RIGHT", frame, "LEFT", -5, 0);
                frame.icon:SetHeight(self.db.profile[frame.configName]["Height"] + 5);
                frame.icon:SetWidth(self.db.profile[frame.configName]["Height"] + 5);
            end
            frame.icon:Show();
        else
            frame.icon:Hide();
        end
    end
    if ((unit == "player") and (frame.unit == "player") and frame.latency and self.db.profile[frame.configName]["ShowLatency"] and frame.sentTime) then
        local min, max = frame:GetMinMaxValues();
        local latency = (GetTime() - frame.sentTime) / (max - min);
        frame.sentTime = nil;
        if (latency < 0) then latency = 0 elseif (latency > 1) then latency = 1 end
        frame.latency:SetWidth(frame:GetWidth() * latency);
        frame.latency:ClearAllPoints();
        if (frame.channeling) then
            frame.latency:SetPoint("LEFT", frame, "LEFT", 0, 0);
        else
            frame.latency:SetPoint("RIGHT", frame, "RIGHT", 0, 0);
        end
        frame.latency:Show();
    end
    if ((unit == "player") and (frame.unit == "player")) then
        local spellName = UnitCastingInfo(unit) or UnitChannelInfo(unit);
        if (spellName) then
            if (self.db.profile[frame.configName]["ShowSpellTarget"]) then
                if (spellName == frame.spellName and frame.spellTargetName and frame.spellTargetName ~= "") then
                    frame.text:SetFormattedText("%s (%s)", frame.spellName, frame.spellTargetName);
                elseif (spellName ~= frame.spellName and UnitName("target")) then
                    frame.text:SetFormattedText("%s (%s)", spellName, UnitName("target"));
                else
                    frame.text:SetText(spellName);
                end
            else
                frame.text:SetText(spellName);
            end
        end
    end
end

function Castbars:FrameLayoutRestore(frame)
    -- Position
    local position = self.db.profile[frame.configName]["Position"];
    if (frame.dragable and position) then
        frame:clearAllPoints();
        frame:setPoint(position.point, position.parent, position.relpoint, position.x, position.y);
    end

    -- Texture
    self:FrameMediaRestore(frame);

    -- Dimensions
    frame:SetWidth(self.db.profile[frame.configName]["Width"]);
    frame:SetHeight(self.db.profile[frame.configName]["Height"]);

    -- Restore dynamic settings
    self:FrameRestoreOnCast(frame);

    -- Visibility
    if (self.db.profile[frame.configName]["Show"]) then
        if (frame == PetCastingBarFrame) then
            frame.showCastbar = self.db.profile[frame.configName]["Show"] or UnitIsPossessed("pet");
        else
            frame.showCastbar = true;
        end
    else
        frame.showCastbar = false;
    end

    -- Shield visibility
    if (frame.shield) then
        if (self.db.profile[frame.configName]["ShowShield"]) then
            frame.shield:SetTexture("Interface\\CastingBar\\UI-CastingBar-Small-Shield");
            frame.shield:SetGradientAlpha("HORIZONTAL", 1, 1, 1, 0.5, 1, 1, 1, 0.5);
        else
            frame.shield:SetTexture();
        end
    end

    -- Total cast time
    if (frame.timer) then
        if (self.db.profile[frame.configName]["ShowTotalCastTime"]) then
            frame.castTimeFormat = "%.1f/%.1f";
        else
            frame.castTimeFormat = "%.1f";
        end
        frame.timer:SetFormattedText(frame.castTimeFormat, 0, 0);
    end

    -- Text Alignment
    frame.text:SetJustifyH(self.db.profile[frame.configName]["TextAlignment"]);
end

function Castbars:FrameLayoutRestoreAll()
    for i, frame in pairs(self.frames) do
        self:FrameLayoutRestore(frame)
    end
end

function Castbars:FrameCustomize(frame)
    local frameName = frame:GetName();
    local frameType;

    if (frameName:sub(1, 11) == "MirrorTimer") then
        local index = tonumber(frameName:sub(12,12));
        frameType = "Mirror";
        frame.statusBar = _G[frameName .. "StatusBar"];
        frame.configName = "MirrorTimer";
        frame.friendlyName = "Mirror Timer " .. index;
        if (index == 1) then
            frame.dragable = true;
        end
    elseif (frameName == "CastingBarFrame") then
        frameType = "Castbar";
        frame.statusBar = frame;
        frame.configName = frameName;
        frame.friendlyName = "Player/Vehicle Castbar";
        frame.dragable = true;
        frame.icon = _G[frameName .. "Icon"];
        frame.shield = _G[frameName .. "BorderShield"];
    elseif (frameName == "PetCastingBarFrame") then
        frameType = "Castbar";
        frame.statusBar = frame;
        frame.configName = UnitIsPossessed("pet") and "CastingBarFrame" or "PetCastingBarFrame";
        frame.friendlyName = "Pet Castbar";
        frame.dragable = true;
        frame.icon = _G[frameName .. "Icon"];
        frame.shield = _G[frameName .. "BorderShield"];
    elseif (frameName == "TargetCastingBarFrame") then
        frameType = "Castbar";
        frame.statusBar = frame;
        frame.configName = frameName;
        frame.friendlyName = "Target Castbar";
        frame.dragable = true;
        frame.icon = _G[frameName .. "Icon"];
        frame.shield = _G[frameName .. "BorderShield"];
    elseif (frameName == "FocusCastingBarFrame") then
        frameType = "Castbar";
        frame.statusBar = frame;
        frame.configName = frameName;
        frame.friendlyName = "Focus Castbar";
        frame.dragable = true;
        frame.icon = _G[frameName .. "Icon"];
        frame.shield = _G[frameName .. "BorderShield"];
    end

    -- Make dragable
    if (frame.dragable) then
        frame:SetMovable(true);
        frame:RegisterForDrag("LeftButton");
        frame:SetScript("OnDragStart", function(frame)
            if (self.ConfigMode) then
                frame:StartMoving();
            end
        end);
        frame:SetScript("OnDragStop", function(frame)
            frame:StopMovingOrSizing();
            if (not self.db.profile[frame.configName]["Position"]) then
                self.db.profile[frame.configName]["Position"] = {};
            end
            local position = self.db.profile[frame.configName]["Position"];
            position.point, position.parent, position.relpoint, position.x, position.y = frame:GetPoint();
        end);
    end

    -- Adjust spark position
    frame.spark = _G[frameName .. "Spark"];
    if (frame.spark) then
        local setPoint = frame.spark.SetPoint;
        frame.spark.SetPoint = function(self, point, relativeFrame, relativePoint, x, y)
            setPoint(self, point, relativeFrame, relativePoint, x, 0);
        end
    end

    -- Adjust text position
    frame.text = _G[frameName .. "Text"];
    frame.text:ClearAllPoints();
    if (frameType == "Castbar") then
        frame.text:SetPoint("LEFT", frame.statusBar, "LEFT", 5, 1);
    elseif (frameType == "Mirror") then
        frame.text:SetPoint("CENTER", frame.statusBar, "CENTER", 0, 1);
    end

    -- Prevent other addons moving the text (and setting the width)
    local frameTextSetWidth = frame.text.SetWidth;
    frame.text.SetWidth = self.DoNothing;
    frame.text.ClearAllPoints = self.DoNothing;
    frame.text.SetPoint = self.DoNothing;

    -- Prevent other addons moving the frame
    frame.clearAllPoints = frame.ClearAllPoints;
    frame.ClearAllPoints = self.DoNothing;
    frame.setPoint = frame.SetPoint;
    frame.SetPoint = self.DoNothing;

    -- Prevent other addons from modifying event registrations
    frame.UnregisterEvent = self.DoNothing;
    frame.UnregisterAllEvents = self.DoNothing;
    frame.RegisterEvent = self.DoNothing;

    -- Remove the border
    local frameBorder = _G[frameName .. "Border"];
    frameBorder:SetTexture();

    -- Remove the flash
    local frameFlash = _G[frameName .. "Flash"];
    if (frameFlash) then
        frameFlash:SetTexture();
    end

    -- Create backdrop
    frame.backdrop = CreateFrame("Frame", frameName .. "Backdrop", frame);
    if (frameType == "Castbar") then
        frame.backdrop:SetPoint("CENTER", frame, "CENTER", 0, 0);
    elseif (frameType == "Mirror") then
        frame.backdrop:SetPoint("CENTER", frame, "CENTER", 0, 3);
    end

    if (frameName == "CastingBarFrame") then
        frame.gcd = CreateFrame("Frame", nil, UIParent);
        frame.gcd:SetPoint("BOTTOM", frame.statusBar, "TOP", 0, 0);
        frame.gcd:SetHeight(4);
        local border = frame.gcd:CreateTexture(nil, "BACKGROUND");
        border:SetAllPoints(frame.gcd);
        local texture = frame.gcd:CreateTexture(nil, "OVERLAY");
        texture:SetTexture("Spells\\AURA_01");
        texture:SetVertexColor(1, 1, 1, 1)
        texture:SetBlendMode('ADD');
        texture:SetWidth(35);
        texture:SetHeight(35);
        frame.gcd:SetScript("OnUpdate", function(frameGcd)
            frameGcd.spellName = frameGcd.spellName or self:GcdSpellNameFind();
            if (not frameGcd.spellName) then return end
            local startTime, duration = GetSpellCooldown(frameGcd.spellName);
            if (duration > 0) then
                texture:SetPoint("CENTER", frameGcd, "LEFT", ((GetTime() - startTime) / duration) * frameGcd:GetWidth(), 0)
                if (frame:IsVisible()) then border:SetTexture(0, 0, 0, 0) else border:SetTexture(0, 0, 0, 0.3) end
            else
                frameGcd:Hide();
            end
        end);
    end

    if (frameType == "Castbar") then
        -- Add timer text
        frame.timer = frame:CreateFontString(nil, "ARTWORK", "GameFontHighlight");
        frame.timer:SetPoint("RIGHT", frame, "RIGHT", -5, 1);
        frame.nextupdate = 0.1;

        -- Add latency indication texture
        if (frameName == 'CastingBarFrame') then
            frame.latency = frame:CreateTexture(nil, "ARTWORK");
            frame.latency:Hide();
            frame.latency:SetHeight(frame:GetHeight());
            frame.latency:SetPoint("RIGHT", frame, "RIGHT", 0, 0);
            frame.latency:SetTexture(1, 0, 0, 0.65);
        end

        -- Adjust shield layer
        frame.shield:SetDrawLayer("OVERLAY");
    elseif (frameType == "Mirror") then
        for i, region in pairs({frame:GetRegions()}) do
            if (region.GetTexture and region:GetTexture() == "SolidTexture") then
                frame.shade = region;
            end
        end
    end

    -- Automatically adjust the height of sub elements
    local setHeight = frame.SetHeight;
    if (frameType == "Castbar") then
        frame.SetHeight = function(self, height)
            frame.backdrop:SetHeight(height + 10);
            frame.spark:SetHeight(2.5 * height);
            if (frame.latency) then
                frame.latency:SetHeight(height);
            end
            frame.shield:SetHeight(5.82 * height);
            setHeight(self, height);
        end
    elseif (frameType == "Mirror") then
        frame.SetHeight = function(self, height)
            frame.backdrop:SetHeight(height + 10);
            frame.statusBar:SetHeight(height);
            frame.shade:SetHeight(height);
            setHeight(self, height + 10);
        end
    end

    -- Automatically adjust the width of sub elements
    local setWidth = frame.SetWidth
    if (frameType == "Castbar") then
        frame.SetWidth = function(self, width)
            frame.backdrop:SetWidth(width + 10);
            frame.timer:SetText("0.0");
            frameTextSetWidth(frame.text, width - 10 - frame.timer:GetWidth());
            frame.shield:SetWidth(1.36 * width);
            frame.shield:ClearAllPoints();
            frame.shield:SetPoint("CENTER", frame.statusBar, "CENTER", -0.031875 * width, 0);
            if (frame.gcd) then
                frame.gcd:SetWidth(width);
            end
            setWidth(self, width);
        end
    elseif (frameType == "Mirror") then
        frame.SetWidth = function(self, width)
            frame.backdrop:SetWidth(width + 10);
            frame.statusBar:SetWidth(width);
            frame.shade:SetWidth(width);
            setWidth(self, width + 10);
        end
    end

    frame:Hide();
end

function Castbars:FrameCustomizeAll()
    for i, frame in pairs(self.frames) do
        self:FrameCustomize(frame)
    end
end

function Castbars:GetOptionsTableForBar(frameConfigName, friendlyName, order, textAlign, showHideShield, enableDisable, showHideIcon, showHideLatency, showHideSpellTarget, showHideTotalCastTime)
    local options =
    {
        type = "group",
        name = friendlyName .. " Options",
        guiInline = true,
        order = order,
        args =
        {
            width =
            {
                name = "Width",
                type = "range",
                order = 10,
                desc = "Set the width of the " .. friendlyName,
                min = 100,
                max = 600,
                step = 1,
                get = function() return self.db.profile[frameConfigName]["Width"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["Width"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
            height =
            {
                name = "Height",
                type = "range",
                order = 11,
                desc = "Set the height of the " .. friendlyName,
                min = 10,
                max = 100,
                step = 1,
                get = function() return self.db.profile[frameConfigName]["Height"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["Height"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
            texture =
            {
                name = "Texture",
                type = "select",
                order = 12,
                desc = "Select texture to use for the " .. friendlyName,
                dialogControl = "LSM30_Statusbar",
                values = AceGUIWidgetLSMlists.statusbar,
                get = function() return self.db.profile[frameConfigName]["Texture"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["Texture"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
            color =
            {
                name = "Custom Bar Color",
                type = "color",
                order = 13,
                desc = "Color of the Castbar.",
                get = function(info)
                    return unpack(self.db.profile[frameConfigName]["BarColor"]);
                end,
                set = function(info, r, g, b)
                    self.db.profile[frameConfigName]["BarColor"] = {r, g, b};
                end,
            },
            customcolor =
            {
                name = "Use Custom Bar Color",
                type = "toggle",
                order = 14,
                desc = "Use the custom bar color selected in the color box",
                get = function() return self.db.profile[frameConfigName]["UseBarColor"] end,
                set = function()
                    self.db.profile[frameConfigName]["UseBarColor"] = not self.db.profile[frameConfigName]["UseBarColor"];
                    self:FrameLayoutRestoreAll();
                end,
            },
            font =
            {
                name = "Font",
                type = "select",
                order = 15,
                desc = "Select font to use for the " .. friendlyName,
                dialogControl = "LSM30_Font",
                values = AceGUIWidgetLSMlists.font,
                get = function() return self.db.profile[frameConfigName]["Font"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["Font"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
            fontsize =
            {
                name = "Font Size",
                type = "range",
                order = 16,
                desc = "Set the font size of the " .. friendlyName,
                min = 6,
                max = 30,
                step = 1,
                get = function() return self.db.profile[frameConfigName]["FontSize"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["FontSize"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
            fontoutline =
            {
                name = "Font Outline",
                type = "toggle",
                order = 17,
                desc = "Toggles outline on the font of the " .. friendlyName,
                get = function() return self.db.profile[frameConfigName]["FontOutline"] end,
                set = function()
                    self.db.profile[frameConfigName]["FontOutline"] = not self.db.profile[frameConfigName]["FontOutline"];
                    self:FrameLayoutRestoreAll();
                end,
            },
            border =
            {
                name = "Border",
                type = "select",
                order = 20,
                desc = "Select border to use for the " .. friendlyName,
                dialogControl = "LSM30_Border",
                values = AceGUIWidgetLSMlists.border,
                get = function() return self.db.profile[frameConfigName]["Border"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["Border"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
            borderbrightness =
            {
                name = "Border Brightness",
                type = "range",
                order = 21,
                desc = "Set the brightness of the border of the " .. friendlyName,
                min = 0.0,
                max = 1.0,
                step = 0.05,
                isPercent = true,
                get = function() return self.db.profile[frameConfigName]["BorderBrightness"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["BorderBrightness"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
        },
    };
    if (enableDisable) then
        options.args.enable =
        {
            name = "Enable",
            type = "toggle",
            order = 1,
            desc = "Toggles display of the " .. friendlyName,
            get = function() return self.db.profile[frameConfigName]["Show"] end,
            set = function()
                self.db.profile[frameConfigName]["Show"] = not self.db.profile[frameConfigName]["Show"];
                self:FrameLayoutRestoreAll();
            end,
        }
    end
    if (showHideIcon) then
        options.args.showicon =
        {
            name = "Show Icon",
            type = "toggle",
            order = 2,
            desc = "Toggles display of the icon at the left side of the bar",
            get = function() return self.db.profile[frameConfigName]["ShowIcon"] end,
            set = function()
                self.db.profile[frameConfigName]["ShowIcon"] = not self.db.profile[frameConfigName]["ShowIcon"];
                self:FrameLayoutRestoreAll();
            end,
        }
    end
    if (showHideShield) then
        options.args.showshield =
        {
            name = "Show Shield",
            type = "toggle",
            order = 3,
            desc = "Toggles display of the shield around the bar when the spell cannot be interrupted.",
            get = function() return self.db.profile[frameConfigName]["ShowShield"] end,
            set = function()
                self.db.profile[frameConfigName]["ShowShield"] = not self.db.profile[frameConfigName]["ShowShield"];
                self:FrameLayoutRestoreAll();
            end,
        }
    end
    if (showHideLatency) then
        options.args.showlatency =
        {
            name = "Show Latency",
            type = "toggle",
            order = 4,
            desc = "Toggles the latency indicator, which shows the latency at the time of spell cast as a red bar at the end of the Castbar.",
            get = function() return self.db.profile[frameConfigName]["ShowLatency"] end,
            set = function()
                self.db.profile[frameConfigName]["ShowLatency"] = not self.db.profile[frameConfigName]["ShowLatency"];
                self:FrameLayoutRestoreAll();
            end,
        }
    end
    if (showHideSpellTarget) then
        options.args.showspelltarget =
        {
            name = "Show Spell Target",
            type = "toggle",
            order = 5,
            desc = "Toggles display of the target of the spell being cast.",
            get = function() return self.db.profile[frameConfigName]["ShowSpellTarget"] end,
            set = function()
                self.db.profile[frameConfigName]["ShowSpellTarget"] = not self.db.profile[frameConfigName]["ShowSpellTarget"];
                self:FrameLayoutRestoreAll();
            end,
        }
    end
    if (showHideTotalCastTime) then
        options.args.showtotalcasttime =
        {
            name = "Show Total Casttime",
            type = "toggle",
            order = 6,
            desc = "Toggles display of the total cast time.",
            get = function() return self.db.profile[frameConfigName]["ShowTotalCastTime"] end,
            set = function()
                self.db.profile[frameConfigName]["ShowTotalCastTime"] = not self.db.profile[frameConfigName]["ShowTotalCastTime"];
                self:FrameLayoutRestoreAll();
            end,
        }
    end
    if (textAlign) then
        options.args.textalign =
        {
            name = "Text Alignment",
            type = "select",
            order = 16,
            desc = "Set the alignment of the Castbar text",
            values = {LEFT = "Left", CENTER = "Center"},
            get = function() return self.db.profile[frameConfigName]["TextAlignment"] end,
            set = function(info, value)
                self.db.profile[frameConfigName]["TextAlignment"] = value;
                self:FrameLayoutRestoreAll();
            end,
        }
    end
    return options;
end

function Castbars:GetOptionsTable()
    local options =
    {
        type = "group",
        args =
        {
            general =
            {
                type = "group",
                name = "General Settings",
                handler = self,
                childGroups = "tree",
                args =
                {
                    general =
                    {
                        type = "group",
                        name = "General Settings",
                        guiInline = true,
                        order = 0,
                        args =
                        {
                            toggleconfigmode =
                            {
                                name = "Configuration Mode",
                                type = "toggle",
                                order = 1,
                                desc = "Toggle configuration mode to allow moving bars and setting appearance options",
                                get = function() return self.ConfigMode end,
                                set = function() self:Toggle() end,
                            },
                        },
                    },
                    player = self:GetOptionsTableForBar("CastingBarFrame", "Player/Vehicle Castbar", 1, true, false, true, true, true, true, true),
                    pet = self:GetOptionsTableForBar("PetCastingBarFrame", "Pet Castbar", 2, true, false, true, true, false, false, true),
                    target = self:GetOptionsTableForBar("TargetCastingBarFrame", "Target Castbar", 3, true, true, true, true, false, false, true),
                    focus = self:GetOptionsTableForBar("FocusCastingBarFrame", "Focus Castbar", 4, true, true, true, true, false, false, true),
                    mirror = self:GetOptionsTableForBar("MirrorTimer", "Mirror Timer Bars", 5),
                },
            },
            profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db),
        },
    };
    return options;
end

function Castbars:UnitFullName(unit)
    local name, realm = UnitName(unit);
    if (realm and realm ~= "") then
        return name .. "-" .. realm;
    else
        return name;
    end
end

function Castbars:NameToUnitID(targetName)
    return UnitExists(targetName) and targetName or (targetName == self:UnitFullName("target") and "target") or (targetName == self:UnitFullName("focus") and "focus") or (targetName == self:UnitFullName("targettarget") and "targettarget") or (targetName == self:UnitFullName("focustarget") and "focustarget") or nil;
end

function Castbars:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("CastbarsDB",
    {
        profile =
        {
            ["**"] =
            {
                Width = 205,
                Height = 13,
                Texture = "Blizzard",
                Border = "Blizzard Tooltip",
                Font = "Friz Quadrata TT",
                FontSize = 12,
                FontOutline = true,
                TextAlignment = "CENTER",
                BorderBrightness = 0.2,
                Position = nil,
                BarColor = {0.3, 0.3, 0.8},
                UseBarColor = true,
            },
            CastingBarFrame =
            {
                Show = true,
                ShowLatency = true,
                ShowIcon = true,
                ShowSpellTarget = true,
                ShowTotalCastTime = true,
                Width = 240,
                Height = 24,
                Position = {point = "BOTTOM", relpoint = "BOTTOM", x = 0, y = 220},
            },
            PetCastingBarFrame =
            {
                Show = true,
                ShowIcon = true,
                Width = 150,
                Height = 12,
                Border = "None",
                Position = {point = "BOTTOM", relpoint = "BOTTOM", x = 0, y = 200},
            },
            TargetCastingBarFrame =
            {
                Show = true,
                ShowIcon = true,
                ShowShield = true,
                Height = 12,
                Border = "None",
                Position = {point = "CENTER", relpoint = "CENTER", x = 0, y = 200},
            },
            FocusCastingBarFrame =
            {
                Show = false,
                ShowIcon = true,
                ShowShield = true,
                Height = 12,
                Border = "None",
                Position = {point = "CENTER", relpoint = "CENTER", x = 0, y = 220},
            },
            MirrorTimer =
            {
            },
        }
    });
    self.db.RegisterCallback(self, "OnProfileChanged", "FrameLayoutRestoreAll");
    self.db.RegisterCallback(self, "OnProfileCopied", "FrameLayoutRestoreAll");
    self.db.RegisterCallback(self, "OnProfileReset", "FrameLayoutRestoreAll");

    self.options = self:GetOptionsTable();
    LibStub("AceConfig-3.0"):RegisterOptionsTable("Castbars", self.options);
    self.BlizzardOptionsFrameGeneral = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Castbars", nil, nil, "general");
    self.BlizzardOptionsFrameProfile = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Castbars", "Profiles", "Castbars", "profile");

    -- Prevent the UIParent from moving the CastingBarFrame around
    UIPARENT_MANAGED_FRAME_POSITIONS["CastingBarFrame"] = nil;

    -- Create target casting bar
    CreateFrame("StatusBar", "TargetCastingBarFrame", UIParent, "CastingBarFrameTemplate");
    TargetCastingBarFrame:RegisterEvent("PLAYER_TARGET_CHANGED");
    CastingBarFrame_OnLoad(TargetCastingBarFrame, "target", false, true);

    -- Create focus casting bar
    CreateFrame("StatusBar", "FocusCastingBarFrame", UIParent, "CastingBarFrameTemplate");
    FocusCastingBarFrame:RegisterEvent("PLAYER_FOCUS_CHANGED");
    CastingBarFrame_OnLoad(FocusCastingBarFrame, "focus", false, true);

    -- Register UNIT_SPELLCAST_SENT on CastingBarFrame
    CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_SENT");

    -- Setup table with all frames
    self.frames = {CastingBarFrame, PetCastingBarFrame, TargetCastingBarFrame, FocusCastingBarFrame, MirrorTimer1, MirrorTimer2, MirrorTimer3};

    -- Customize the bars
    self:FrameCustomizeAll();

    -- Restore layout of all frames
    local loginRestoreOnce;
    if (IsLoggedIn()) then
        self:FrameLayoutRestoreAll();
        loginRestoreOnce = true;
    end

    -- Listen to LibSharedMedia-3.0 callbacks
    self.SharedMedia.RegisterCallback(self, "LibSharedMedia_Registered", "FrameMediaRestoreAll")
    self.SharedMedia.RegisterCallback(self, "LibSharedMedia_SetGlobal", "FrameMediaRestoreAll")

    -- Hook OnUpdate
    local orgCastingBarFrame_OnUpdate = CastingBarFrame_OnUpdate;
    self.CastingBarFrame_OnUpdate = function(frame, elapsed, ...)
        orgCastingBarFrame_OnUpdate(frame, elapsed, ...);
        if (frame:IsVisible()) then
            if (frame.outOfRange ~= nil) then
                local outOfRange = frame.outOfRange;
                frame.outOfRange = frame.spellName and frame.spellTarget and (IsSpellInRange(frame.spellName, frame.spellTarget) == 0) and true or false;
                if (outOfRange ~= frame.outOfRange) then
                    self:FrameRestoreColor(frame);
                end
            end
            if (frame.timer) then
                if (frame.nextupdate < elapsed) then
                    if (frame.casting) then
                        frame.timer:SetFormattedText(frame.castTimeFormat, max(frame.maxValue - frame.value, 0), frame.maxValue);
                    elseif (frame.channeling) then
                        frame.timer:SetFormattedText(frame.castTimeFormat, max(frame.value, 0), frame.maxValue);
                    elseif (self.ConfigMode) then
                        frame.timer:SetFormattedText(frame.castTimeFormat, 0, 0);
                    else
                        frame.timer:SetText();
                    end;
                    frame.nextupdate = 0.1;
                else
                    frame.nextupdate = frame.nextupdate - elapsed;
                end
            end
        end
    end
    CastingBarFrame_OnUpdate = self.CastingBarFrame_OnUpdate;

    -- Hook OnEvent
    local orgCastingBarFrame_OnEvent = CastingBarFrame_OnEvent;
    self.CastingBarFrame_OnEvent = function(frame, event, ...)
        frame.Show = nil; -- Protect against addons that override the Show method
        if (event == "UNIT_SPELLCAST_SENT") then
            local unit, spellName, _, targetName = ...;
            if (unit == "player") then
                frame.sentTime = GetTime();
                frame.outOfRange = false;
                frame.spellName = spellName;
                frame.spellTargetName = targetName;
                frame.spellTarget = self:NameToUnitID(targetName);
                frame.gcd:SetFrameLevel(frame.backdrop:GetFrameLevel() + 1);
                frame.gcd:Show();
            end
            return;
        elseif (event == "PLAYER_ENTERING_WORLD") then
            if (not loginRestoreOnce) then
                self:FrameLayoutRestoreAll();
                loginRestoreOnce = true;
            end
        elseif (event == "PLAYER_TARGET_CHANGED") then
            event = "PLAYER_ENTERING_WORLD";
        elseif ((event == "UNIT_PET") and (... == "player") and (frame == PetCastingBarFrame)) then
            frame.configName = UnitIsPossessed("pet") and "CastingBarFrame" or "PetCastingBarFrame";
            self:FrameLayoutRestore(frame);
            return;
        end
        orgCastingBarFrame_OnEvent(frame, event, ...);
        if ((event == "UNIT_SPELLCAST_START") or (event == "UNIT_SPELLCAST_CHANNEL_START")) then
            if (frame.statusBar and frame:IsVisible()) then
                local unit = ...;
                self:FrameRestoreOnCast(frame, unit);
            end
        elseif ((event == "UNIT_SPELLCAST_STOP") or (event == "UNIT_SPELLCAST_CHANNEL_STOP")) then
            if (frame.latency and not (frame.casting or frame.channeling)) then
                frame.latency:Hide();
            end
        end
    end
    CastingBarFrame_OnEvent = self.CastingBarFrame_OnEvent;

    -- Bypass the PetCastingBarFrame_OnEvent function for the PetCastingBarFrame
    PetCastingBarFrame:SetScript("OnEvent", self.CastingBarFrame_OnEvent);

    self.GetOptionsTableForBar = nil;
    self.GetOptionsTable = nil;
    self.FrameCustomize = nil;
    self.FrameCustomizeAll = nil;
    self.OnInitialize = nil;
end

--[[ ConfigMode Support ]]--

Castbars.MirrorTimerFrame_OnEvent = MirrorTimerFrame_OnEvent;
Castbars.MirrorTimer_Show = MirrorTimer_Show;

function Castbars:Show()
    if (not self.ConfigMode) then
        self.ConfigMode = true;

        -- Prevent event handling
        CastingBarFrame_OnEvent = self.DoNothing;
        MirrorTimerFrame_OnEvent = self.DoNothing
        MirrorTimer_Show = self.DoNothing;

        -- Make the Pet Castbar appear as a Pet Castbar even if in a vehicle while configuring.
        PetCastingBarFrame.configName = "PetCastingBarFrame";
        self:FrameLayoutRestore(PetCastingBarFrame);

        -- Condition and show all bars
        for i, frame in pairs(self.frames) do
            frame:EnableMouse(true);
            frame.text:SetText(frame.friendlyName);
            frame.statusBar:SetStatusBarColor(unpack(self.db.profile[frame.configName]["BarColor"]));
            frame.statusBar:SetAlpha(1);
            frame.statusBar:SetValue(select(2, frame.statusBar:GetMinMaxValues()));
            if (frame.spark) then
                frame.spark:Hide();
            end
            frame.fadeOut = nil;
            frame.paused = 1;
            frame:Show();
        end
    end
end

function Castbars:Hide()
    if (self.ConfigMode) then
        for i, frame in pairs(self.frames) do
            frame:EnableMouse(false);
            frame:Hide();
        end

        -- Restore the config name of the Pet Castbar
        PetCastingBarFrame.configName = UnitIsPossessed("pet") and "CastingBarFrame" or "PetCastingBarFrame";
        self:FrameLayoutRestore(PetCastingBarFrame);

        -- Restore event handling
        CastingBarFrame_OnEvent = self.CastingBarFrame_OnEvent;
        MirrorTimerFrame_OnEvent = self.MirrorTimerFrame_OnEvent;
        MirrorTimer_Show = self.MirrorTimer_Show;
        self.ConfigMode = false;
    end
end

function Castbars:Toggle()
    if (self.ConfigMode) then
        self:Hide();
    else
        self:Show();
    end
end

CONFIGMODE_CALLBACKS = CONFIGMODE_CALLBACKS or {};
CONFIGMODE_CALLBACKS["Castbars"] = function(action)
    if (action == "ON") then Castbars:Show()
    elseif (action == "OFF") then Castbars:Hide() end
end
